/*
 * DupliphyML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DupliphyML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DupliphyML.  If not, see <http://www.gnu.org/licenses/>.
 */



import Alignments.Alignment;
import Alignments.Ambiguous;
import Alignments.DuplicationAlignment;
import Alignments.Site;
import Ancestors.AncestralJoint;
import Exceptions.GeneralException;
import Exceptions.OutputException;
import Likelihood.Calculator;
import Likelihood.Likelihood;
import Likelihood.Likelihood.SiteLikelihood;
import Maths.SquareMatrix;
import Optimizers.Optimizer;
import Parameters.Parameter;
import Parameters.Parameters;
import Models.Distributions;
import Models.RateCategory;
import Models.Model;
import Optimizers.GoldenSection;
import Trees.Tree;
import Utils.SettingException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;

/**
 * DupliphyML code.  This is pretty much a simple wrapper round GeLL
 * @author Daniel Money
 * @version 1.1
 */
public class DupliphyML
{
    /**
     * The main method.  See documentation for what command line inputs are
     * required.
     * @param args Command line arguements
     */
    public static void main(String[] args)
    {
	try
	{
	    if ((args.length < 5) || (args.length > 7))
	    {
		throw new SettingException("Wrong number of command line inputs");
	    }
	    Distributions.setMethod(Distributions.Calculation.REPEAT);
	    SquareMatrix.setExpMethod(SquareMatrix.Calculation.TAYLOR);

	    Alignment a = new DuplicationAlignment(new File(args[0]));
	    Tree t = Tree.fromFile(new File(args[2]));

	    RateCategory sr;
	    Model r;
	    Parameters p;

	    int num = Integer.parseInt(args[4]);

	    int f= 0;
	    double n = num;
	    while (n > 12)
	    {
		n = n / 12;
		f++;
	    }
	    SquareMatrix.setForce(f);

	    Alignment missing = null;
	    if (args.length == 7)
	    {
		Ambiguous ambig = Ambiguous.fromFile(new File(args[6]));
		missing = new DuplicationAlignment(new File(args[5]),ambig);
	    }
	    if (args.length == 6)
	    {
		missing = new DuplicationAlignment(new File(args[5]));
	    }

            //Doesn'ty use the factory class of GePuL as that was written later
            //than this!
	    String[][] matrix = new String[num+1][num+1];
	    HashMap<String,Integer> map = new HashMap<String,Integer>();
	    for (int i = 0; i <= num; i++)
	    {
		map.put(Integer.toString(i), i);
	    }
	    if (args[1].contains("Parsimony"))
	    {
		for (int i = 0; i <= num; i++)
		{
		    for (int j = 0; j <= num; j++)
		    {
			if (Math.abs(i-j) == 1)
			{
			    matrix[i][j] = "1.0";
			}
			else
			{
			    matrix[i][j] = "0.0";
			}
		    }
		}

		sr = new RateCategory(matrix,RateCategory.FrequencyType.STATIONARY,map);
		p = new Parameters();
	    }
	    else
	    {
		for (int i = 0; i <= num; i++)
		{
		    for (int j = 0; j <= num; j++)
		    {
			matrix[i][j] = "0.0";
			if (i-j == 1)
			{
			    matrix[i][j] = "d";
			}
			if (j-i == 1)
			{
			    matrix[i][j] = "b";
			}
			if ((i == 0) && (j == 1))
			{
			    matrix[i][j] = "i";
			}
		    }
		}

		sr = new RateCategory(matrix,RateCategory.FrequencyType.STATIONARY,map);

		p = new Parameters();
		p.addParameter(Parameter.newFixedParameter("b", 1.0));
                //Bound parameters so they can't go to zero as this would create a sink state model
		p.addParameter(Parameter.newEstimatedBoundedParameter("d",1e-4,Double.MAX_VALUE));
                //Further bound the innovation parameter, otherwise can go to infinity
                //on some simulated datasets with very little change.  This is due
                //to there being zero probability of being in state zero at the root
                //and very little chance of changing to zero over the tree.  If the
                //simulated dataset has no zero innovation will go to infinity if
                //not bound.
		p.addParameter(Parameter.newEstimatedBoundedParameter("i",1e-4,10.0));
	    }
	    if (args[1].contains("+G"))
	    {
		r = Model.gammaRates(sr, "g", 4);
		p.addParameter(Parameter.newEstimatedBoundedParameter("g", 0.2, 10.0));
	    }
	    else
	    {
		r = new Model(sr);
	    }
	    p.addParameters(t.getParametersForEstimation());

	    Calculator l = new Calculator(r,a,t,missing);
	    Optimizer o = new GoldenSection(GoldenSection.ProgressLevel.PARAMETER);
            //Replace the above with the line below to use Neader-Mead optimization
            //ptimizer o = new NelderMead();

	    Likelihood like = o.maximise(l, p);

            //Get the optimized parameters
            Parameters op = like.getParameters();
            
	    Tree lt = new Tree(t,op);

	    AncestralJoint aj = AncestralJoint.newInstance(r,a,lt);
	    Alignment anc = aj.calculate(op);

	    op.toFile(new File(args[3] + "_params.txt"), like.getLikelihood());
	    lt.toFile(new File(args[3] + "_tree.ph"),true);
	    DuplicationAlignment.writeFile(anc, new File(args[3] + "_desc.txt"));

            if (args[1].contains("+G"))
	    {
                File fr = new File(args[3] + "_rate.txt");
                try
                {
                    PrintStream rout = new PrintStream(new FileOutputStream(fr));
                    for (int i = 0; i < a.getLength(); i++)
                    {    
                        Site s = a.getSite(i);
                        SiteLikelihood sl = like.getSiteLikelihood(s);
                        double rate = 0.0;
                        for (RateCategory rc: r)
                        {
                            rate += rc.getTotalRate() * sl.getRateCategoryProbability(rc) * r.getScale();
                        }
                        rout.println(i + "\t" + rate);
                    }
                    rout.close();
                }
                catch (FileNotFoundException e)
                {
                    throw new OutputException(fr.getAbsolutePath(),"Unable to write out duplication alignment",e);
                }
            }
	}
	catch (GeneralException e)
	{
	    System.err.println(e.toString());
	}
	catch (SettingException e)
	{
	    System.err.println(e.toString());
	}

    }
}
